package api

import (
	kyvernov1 "github.com/kyverno/kyverno/api/kyverno/v1"
	policiesv1alpha1 "github.com/kyverno/kyverno/api/policies.kyverno.io/v1alpha1"
	admissionregistrationv1 "k8s.io/api/admissionregistration/v1"
	admissionregistrationv1alpha1 "k8s.io/api/admissionregistration/v1alpha1"
	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
)

// Everything someone might need to validate a single ValidatingAdmissionPolicy
// against all of its registered bindings.
type ValidatingAdmissionPolicyData struct {
	definition *admissionregistrationv1.ValidatingAdmissionPolicy
	bindings   []admissionregistrationv1.ValidatingAdmissionPolicyBinding
}

func (p *ValidatingAdmissionPolicyData) AddBinding(binding admissionregistrationv1.ValidatingAdmissionPolicyBinding) {
	p.bindings = append(p.bindings, binding)
}

func (p *ValidatingAdmissionPolicyData) GetDefinition() *admissionregistrationv1.ValidatingAdmissionPolicy {
	return p.definition
}

func (p *ValidatingAdmissionPolicyData) GetBindings() []admissionregistrationv1.ValidatingAdmissionPolicyBinding {
	return p.bindings
}

func NewValidatingAdmissionPolicyData(
	policy *admissionregistrationv1.ValidatingAdmissionPolicy,
	bindings ...admissionregistrationv1.ValidatingAdmissionPolicyBinding,
) *ValidatingAdmissionPolicyData {
	return &ValidatingAdmissionPolicyData{
		definition: policy,
		bindings:   bindings,
	}
}

// GenericPolicy abstracts the policy type (ClusterPolicy/Policy, ValidatingPolicy, ValidatingAdmissionPolicy and MutatingAdmissionPolicy)
// It is intended to be used in EngineResponse
type GenericPolicy interface {
	metav1.Object
	// GetAPIVersion returns policy API version
	GetAPIVersion() string
	// GetKind returns policy kind
	GetKind() string
	// IsNamespaced indicates if the policy is namespace scoped
	IsNamespaced() bool
	// AsObject returns the raw underlying policy
	AsObject() any
	// AsKyvernoPolicy returns the kyverno policy
	AsKyvernoPolicy() kyvernov1.PolicyInterface
	// AsValidatingAdmissionPolicy returns the validating admission policy with its bindings
	AsValidatingAdmissionPolicy() *ValidatingAdmissionPolicyData
	// AsValidatingPolicy returns the validating policy
	AsValidatingPolicy() *policiesv1alpha1.ValidatingPolicy
	// AsImageValidatingPolicy returns the imageverificationpolicy
	AsImageValidatingPolicy() *policiesv1alpha1.ImageValidatingPolicy
}

type genericPolicy struct {
	metav1.Object
	PolicyInterface           kyvernov1.PolicyInterface
	ValidatingAdmissionPolicy *ValidatingAdmissionPolicyData
	MutatingAdmissionPolicy   *admissionregistrationv1alpha1.MutatingAdmissionPolicy
	ValidatingPolicy          *policiesv1alpha1.ValidatingPolicy
	ImageValidatingPolicy     *policiesv1alpha1.ImageValidatingPolicy
}

func (p *genericPolicy) AsObject() any {
	return p.Object
}

func (p *genericPolicy) AsKyvernoPolicy() kyvernov1.PolicyInterface {
	return p.PolicyInterface
}

func (p *genericPolicy) AsValidatingAdmissionPolicy() *ValidatingAdmissionPolicyData {
	return p.ValidatingAdmissionPolicy
}

func (p *genericPolicy) AsValidatingPolicy() *policiesv1alpha1.ValidatingPolicy {
	return p.ValidatingPolicy
}

func (p *genericPolicy) AsImageValidatingPolicy() *policiesv1alpha1.ImageValidatingPolicy {
	return p.ImageValidatingPolicy
}

func (p *genericPolicy) GetAPIVersion() string {
	switch {
	case p.PolicyInterface != nil:
		return kyvernov1.GroupVersion.String()
	case p.ValidatingAdmissionPolicy != nil:
		return admissionregistrationv1.SchemeGroupVersion.String()
	case p.MutatingAdmissionPolicy != nil:
		return admissionregistrationv1alpha1.SchemeGroupVersion.String()
	case p.ValidatingPolicy != nil:
		return policiesv1alpha1.GroupVersion.String()
	case p.ImageValidatingPolicy != nil:
		return policiesv1alpha1.GroupVersion.String()
	}
	return ""
}

func (p *genericPolicy) GetKind() string {
	switch {
	case p.PolicyInterface != nil:
		return p.PolicyInterface.GetKind()
	case p.ValidatingAdmissionPolicy != nil:
		return "ValidatingAdmissionPolicy"
	case p.MutatingAdmissionPolicy != nil:
		return "MutatingAdmissionPolicy"
	case p.ValidatingPolicy != nil:
		return "ValidatingPolicy"
	case p.ImageValidatingPolicy != nil:
		return "ImageValidatingPolicy"
	}
	return ""
}

func (p *genericPolicy) IsNamespaced() bool {
	switch {
	case p.PolicyInterface != nil:
		return p.PolicyInterface.IsNamespaced()
	}
	return false
}

func NewKyvernoPolicy(pol kyvernov1.PolicyInterface) GenericPolicy {
	return &genericPolicy{
		Object:          pol,
		PolicyInterface: pol,
	}
}

func NewValidatingAdmissionPolicy(pol *admissionregistrationv1.ValidatingAdmissionPolicy) GenericPolicy {
	return &genericPolicy{
		Object:                    pol,
		ValidatingAdmissionPolicy: NewValidatingAdmissionPolicyData(pol),
	}
}

func NewValidatingAdmissionPolicyWithBindings(pol *admissionregistrationv1.ValidatingAdmissionPolicy, bindings ...admissionregistrationv1.ValidatingAdmissionPolicyBinding) GenericPolicy {
	return &genericPolicy{
		Object:                    pol,
		ValidatingAdmissionPolicy: NewValidatingAdmissionPolicyData(pol, bindings...),
	}
}

func NewMutatingAdmissionPolicy(pol *admissionregistrationv1alpha1.MutatingAdmissionPolicy) GenericPolicy {
	return &genericPolicy{
		Object:                  pol,
		MutatingAdmissionPolicy: pol,
	}
}

func NewValidatingPolicy(pol *policiesv1alpha1.ValidatingPolicy) GenericPolicy {
	return &genericPolicy{
		Object:           pol,
		ValidatingPolicy: pol,
	}
}

func NewImageValidatingPolicy(pol *policiesv1alpha1.ImageValidatingPolicy) GenericPolicy {
	return &genericPolicy{
		Object:                pol,
		ImageValidatingPolicy: pol,
	}
}
